/* 
 *     This file is part of K14, the LCD controller.
 * 
 * Copyright (C) 2014 Dmitry Kobylin
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <debug.h>
#include <LPC177x_8x.h>
#include <LPC177x_8x_bits.h>
#include "command.h"
#include "extspi.h"
#include "gs/gs.h"

//#define DEBUG_COMMAND

#ifdef DEBUG_COMMAND
    #define DPRINT(fmt, ...)  dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif

/* P0.15 */
#define BUSY(value)        do { if (value) LPC_GPIO0->SET  = (1 << 15); \
                                      else LPC_GPIO0->CLR  = (1 << 15); } while (0)
#define BUSY_DIR_OUTPUT()  do {            LPC_GPIO0->DIR |= (1 << 15);} while (0)

/* P2.26 */
#define MISO_BUSY(value)        do { if (value) LPC_GPIO2->SET  = (1 << 26); \
                                      else LPC_GPIO2->CLR  = (1 << 26); } while (0)
#define MISO_BUSY_DIR_OUTPUT()  do {            LPC_GPIO2->DIR |= (1 << 26);} while (0)


#define CMD_FLIP            0x11
#define CMD_DRAW_LINE       0x54
#define CMD_DRAW_RECT       0x55
#define CMD_FILLED_RECT     0x56
#define CMD_ELLIPSE         0x58
#define CMD_FILL_ELLIPSE  0x59

#pragma pack(push, 1)
/*
 *   CMD 0x11 - flip display (swap front/back buffer)
 *
 *       byte | value
 *       ----------------------
 *          0 | 0x11
 *
 *
 *   CMD 0x55 - draw rectangular
 *
 *       byte | value
 *       ----------------------
 *          0 | 0x55
 *          1 | LAYER
 *          2 | X0 (LSB)
 *          3 | X0 (MSB)
 *          4 | Y0 (LSB)
 *          5 | Y0 (MSB)
 *          6 | X1 (LSB)
 *          7 | X1 (MSB)
 *          8 | Y1 (LSB)
 *          9 | Y1 (MSB)
 *         10 | LINE WIDTH
 *         11 | COLOR (LSB)
 *         12 | COLOR (MSB)
 *
 *         LAYER
 *             Layer to draw to:
 *                 0 - front buffer 
 *                 1 - back buffer 
 *                 2 - first layer (not implemented)
 *                 3 - second layer (not implemented)
 */
struct command_data_t {
    uint8 cmd;
    union {
        struct cmd_draw_line {
            uint8     layer;
            uint16    x0;
            uint16    y0;
            uint16    x1;
            uint16    y1;
            GS_COLOR  color;
        } draw_line;
        struct cmd_draw_rect {
            uint8     layer;
            uint16    x0;
            uint16    y0;
            uint16    x1;
            uint16    y1;
            uint8     lwidth;
            GS_COLOR  color;
        } draw_rect;
        struct cmd_filled_rect{
            uint8     layer;
            uint16    x0;
            uint16    y0;
            uint16    x1;
            uint16    y1;
            uint8     dir;  
            GS_COLOR  color;
            GS_COLOR  color1;
        } filled_rect;
        struct cmd_ellipse{
            uint8     layer;
            uint16    xc;
            uint16    yc;
            uint16    rh;
            uint16    rv;
            uint8     ewidth;
            GS_COLOR  color;
        } ellipse;
        struct cmd_fill_ellipse{
            uint8     layer;
            uint16    xc;
            uint16    yc;
            uint16    rh;
            uint16    rv;
            GS_COLOR  color;
        } fill_ellipse;
    };
};
#pragma pack(pop)


struct command_t {
    uint32 rxc; /* number of received bytes */
    enum command_interface_t iface;
    enum {
        COMMAND_BYTE_ORDER_LITTLE_ENDIAN, 
        COMMAND_BYTE_ORDER_BIG_ENDIAN, 
    } byteorder;
};

static struct command_t command;
static struct command_data_t cmddata;

static void _swap_data();

/****************** Command Init *****************/
void command_init()
{
//    MISO_BUSY_DIR_OUTPUT();
//    MISO_BUSY(1);

    command.byteorder = COMMAND_BYTE_ORDER_BIG_ENDIAN;
    extspi_init();
}

/****************** Command Process *************/
void command_process()
{
    extspi_process();
}

/***************** Reset command interface *********/
void command_reset()
{
    command.rxc = 0;
//    MISO_BUSY(1);
}

/****************** Receive byte cb ***************/
void command_receive_cb(uint8 byte)
{
    uint32 rxc;
    uint8 *buf;

    DPRINT("<receive >1x_4dn", byte, command.rxc);

    if (command.rxc < sizeof(struct command_data_t))
            LPC_IOCON_FUNC_SET(P2_26, LPC_IOCON_FUNC_GPIO);
            MISO_BUSY_DIR_OUTPUT();
            MISO_BUSY(0);     // MISO low
    {
        buf = (uint8*)&cmddata;
        buf[command.rxc++] = byte;
        rxc = command.rxc;
        if (rxc >= 1)
        {

            switch (cmddata.cmd)
            {
 /*************************************** FLIP Command  0x11 ***********************************/
               case CMD_FLIP:                                   
                    gs_cmd_flip();    
                    command_reset();            /* we are ready to receive next command */
                    break;

/**************************************** LINE Command 0x54 ***********************************/
                case CMD_DRAW_LINE:                     
                    if (rxc >= 1 + (sizeof(struct cmd_draw_line)))
                    {
                        if (command.byteorder == COMMAND_BYTE_ORDER_BIG_ENDIAN)
                            _swap_data();
                        gs_cmd_line(gs_get_layer(cmddata.draw_line.layer),
                                cmddata.draw_line.x0,
                                cmddata.draw_line.y0,
                                cmddata.draw_line.x1,
                                cmddata.draw_line.y1,
                                cmddata.draw_line.color);

                        command_reset();        /* we are ready to receive next command */
                    }
                    break;

/************************************* RECTANGLE Command 0x55 ********************************/
                case CMD_DRAW_RECT:                     
                    if (rxc >= 1 + (sizeof(struct cmd_draw_rect)))
                    {
                        if (command.byteorder == COMMAND_BYTE_ORDER_BIG_ENDIAN)
                            _swap_data();

                        gs_cmd_rect(gs_get_layer(cmddata.draw_rect.layer),
                                cmddata.draw_rect.x0,
                                cmddata.draw_rect.y0,
                                cmddata.draw_rect.x1,
                                cmddata.draw_rect.y1,
                                cmddata.draw_rect.lwidth,
                                cmddata.draw_rect.color);

                        command_reset();        /* we are ready to receive next command */
                    }
                    break;

/************************************ FILLED RECTANGLE Command 0x56 ****************************/
                case CMD_FILLED_RECT:                   
                    if (rxc >= 1 + (sizeof(struct cmd_filled_rect)))
                    {
                        if (command.byteorder == COMMAND_BYTE_ORDER_BIG_ENDIAN)
                            _swap_data();

                        gs_cmd_filled_rect(gs_get_layer(cmddata.filled_rect.layer),
                                cmddata.filled_rect.x0,
                                cmddata.filled_rect.y0,
                                cmddata.filled_rect.x1,
                                cmddata.filled_rect.y1,
                                cmddata.filled_rect.dir,
                                cmddata.filled_rect.color,
                                cmddata.filled_rect.color1);
                        command_reset();        /* we are ready to receive next command */
                    }
                    break;

/************************************ ELLIPSE Command 0x58 ****************************/
                case CMD_ELLIPSE:                       
                    if (rxc >= 1 + (sizeof(struct cmd_ellipse)))
                    {
                        if (command.byteorder == COMMAND_BYTE_ORDER_BIG_ENDIAN)
                            _swap_data();

                        gs_cmd_ellipse(gs_get_layer(cmddata.ellipse.layer),
                                cmddata.ellipse.xc,
                                cmddata.ellipse.yc,
                                cmddata.ellipse.rh,
                                cmddata.ellipse.rv,
                                cmddata.ellipse.ewidth,
                                cmddata.ellipse.color);

                        command_reset();        /* we are ready to receive next command */
                    }
                    break;

/************************************ FILLED ELLIPSE Command 0x59 ****************************/
                case CMD_FILL_ELLIPSE:                  
                    if (rxc >= 1 + (sizeof(struct cmd_fill_ellipse)))
                    {
                        if (command.byteorder == COMMAND_BYTE_ORDER_BIG_ENDIAN)
                            _swap_data();

                        gs_cmd_fill_ellipse(gs_get_layer(cmddata.fill_ellipse.layer),
                                cmddata.fill_ellipse.xc,
                                cmddata.fill_ellipse.yc,
                                cmddata.fill_ellipse.rh,
                                cmddata.fill_ellipse.rv,
                                cmddata.fill_ellipse.color);

                        command_reset();        /* we are ready to receive next command */
                    }
                    break;

                
                default:
                    DPRINT("<unknown command >1xn", cmddata.cmd);

            }

        }
    }

    //MISO_BUSY(1);    // MISO ligh
    LPC_IOCON_FUNC_SET(P2_26, LPC_IOCON_P2_26_SSP0_MISO);
}

/******************** Command Interface ****************/
void command_set_iface(enum command_interface_t iface)
{
    command.iface = COMMAND_INTERFACE_SPI;
}

/***********/
int command_send_data(uint8 *byte)
{
    return 0;
}

//******************* Swap Bytes MSB-LSB ***************** 
#define SWAP16(val) val = ((val) >> 8) | ((val) << 8)
static void _swap_data()
{
    switch (cmddata.cmd)
    {
        case CMD_DRAW_LINE:
            SWAP16(cmddata.draw_line.x0);
            SWAP16(cmddata.draw_line.y0);
            SWAP16(cmddata.draw_line.x1);
            SWAP16(cmddata.draw_line.y1);
            SWAP16(cmddata.draw_line.color);
            break;
        case CMD_DRAW_RECT:
            SWAP16(cmddata.draw_rect.x0);
            SWAP16(cmddata.draw_rect.y0);
            SWAP16(cmddata.draw_rect.x1);
            SWAP16(cmddata.draw_rect.y1);
            SWAP16(cmddata.draw_rect.color);
            break;
        case CMD_FILLED_RECT:
            SWAP16(cmddata.filled_rect.x0);
            SWAP16(cmddata.filled_rect.y0);
            SWAP16(cmddata.filled_rect.x1);
            SWAP16(cmddata.filled_rect.y1);
            SWAP16(cmddata.filled_rect.color);
            SWAP16(cmddata.filled_rect.color1);
            break;
        case CMD_ELLIPSE:
            SWAP16(cmddata.ellipse.xc);
            SWAP16(cmddata.ellipse.yc);
            SWAP16(cmddata.ellipse.rh);
            SWAP16(cmddata.ellipse.rv);
            SWAP16(cmddata.ellipse.color);
            break;
        case CMD_FILL_ELLIPSE:
            SWAP16(cmddata.fill_ellipse.xc);
            SWAP16(cmddata.fill_ellipse.yc);
            SWAP16(cmddata.fill_ellipse.rh);
            SWAP16(cmddata.fill_ellipse.rv);
            SWAP16(cmddata.fill_ellipse.color);
            break;
    }
}

