/* 
 *     This file is part of K14, the LCD controller.
 * 
 * Copyright (C) 2014 Dmitry Kobylin
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include <debug.h>
#include <LPC177x_8x.h>
#include <LPC177x_8x_bits.h>
#include <clk_cfg.h>
#include "extspi.h"
#include "command.h"

#define DEBUG_EXTSPI

#ifdef DEBUG_EXTSPI
    #define DPRINT(fmt, ...)  dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif


enum extspi_state_t {
    EXTSPI_STATE_IDLE,
    EXTSPI_STATE_WAIT_DATA,
};

static enum extspi_state_t state;

static int _rx_data();


/*
 *
 */
void extspi_init()
{
    /* Setup IO for SPI mode 
     * EXT_SS     P2.23/SSEL0
     * EXT_SCK    P2.22/SCK0
     * EXT_MOSI   P2.26/MISO0
     * EXT_MISO   P2.27/MOSI0
     */
    {
#define CS_VAL() (LPC_GPIO2->PIN & (1 << 23))
        LPC_IOCON_FUNC_SET(P2_23, LPC_IOCON_P2_23_SSP0_SSEL);
        LPC_IOCON_FUNC_SET(P2_22, LPC_IOCON_P2_22_SSP0_SCK);
        LPC_IOCON_FUNC_SET(P2_26, LPC_IOCON_P2_26_SSP0_MISO);
        LPC_IOCON_FUNC_SET(P2_27, LPC_IOCON_P2_27_SSP0_MOSI);
    }
#define EXTSPI_SSP    LPC_SSP0
#define PCONP_BIT     PCONP_PCSSP0

#define GET_DATA()         (EXTSPI_SSP->DR)
#define SEND_DATA(byte)    (EXTSPI_SSP->DR = byte)

    /* Setup SSP controller */
    {
        /*
         * NOTE
         *
         * Clock rate not relevant in slave mode, but setup it anyway.
         *
         * In Slave mode, the SSP clock rate provided by the master must
         * not exceed 1/12 of the peripheral clock selected in
         * Section 3.3.21. The content of the SSPnCPSR register is not
         * relevant.
         *
         *
         * Freq = PCLK / (CPSDVSR * (SCR + 1))
         *
         * speed bits:
         *     0..7    CPSDVSR  (even values between 2 and 254)
         *     8..15   SCR
         */
#if CLK_PCLK == (24 * 1000 * 1000)
    #define SSP_SPEED_12MHZ_24MHZ    (2   | (0 << 8))

    #define SSP_SPEED     SSP_SPEED_12MHZ_24MHZ
#else
    #error Unknown PCLK frequency
#endif

        LPC_SC->PCONP |= PCONP_BIT;

        /*
         * NOTE
         *     CPOL  0
         *     CPHA  1
         */
        EXTSPI_SSP->CR1  = 0;
        EXTSPI_SSP->CR0  = SSP_CR0_CPHA | SSP_CR0_DSS_8BIT | SSP_CR0_FRF_SPI | SSP_CR0_SCR(SSP_SPEED >> 8); 
        EXTSPI_SSP->CPSR = SSP_CPSR_CPSDVSR(SSP_SPEED & 0xff);
        EXTSPI_SSP->CR1  = SSP_CR1_SSE | SSP_CR1_MS_SLAVE;

        GET_DATA();
    }
}

/*
 *
 */
void extspi_process()
{
    uint8 send;

    /* if CS high - return to IDLE state */
    if (CS_VAL() != 0)
    {
        state = EXTSPI_STATE_IDLE;
        return;
    }

    if (state == EXTSPI_STATE_IDLE)
    {
        state = EXTSPI_STATE_WAIT_DATA;

        /* 
         * Reset command state and select SPI as command interface
         */
        command_reset();
        command_set_iface(COMMAND_INTERFACE_SPI);
    }

    /* wait for incoming byte */
    if (!_rx_data())
        return;

    command_receive_cb(GET_DATA());

    if (command_send_data(&send))
    {
        SEND_DATA(send);
    }
}

/*
 *
 */
static int _rx_data()
{
    if (EXTSPI_SSP->SR & SSP_SR_RNE)
        return 1;
    return 0;
}

