/* 
 *     This file is part of K14, the LCD controller.
 * 
 * Copyright (C) 2014 Dmitry Kobylin
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <debug.h>
#include "gs.h"
#include "gs_hw.h"
#include "gs_data.h"

#define GS_LAYER_COUNT   2
static struct gs_layer_t gslayer[GS_LAYER_COUNT];
uint8 gs_buffer[GS_BUFFER_SIZE];

union gs_pixel_t gs_mem[GS_MEM_BUFNUM][GS_BUF_SIZE] __attribute((section("gsmem")));
struct gs_t gs;

/************************* Init Display ***************************/
void gs_init()
{
    gs_hw_init();

    gs.abuf          = 0;
    gs.bgcolor.value = GS_DEFAULT_COLOR;

    gslayer[GS_LAYER_FRONT_BUFFER].x      = 0;
    gslayer[GS_LAYER_FRONT_BUFFER].y      = 0;
    gslayer[GS_LAYER_FRONT_BUFFER].width  = GS_RES_HORIZONTAL;
    gslayer[GS_LAYER_FRONT_BUFFER].height = GS_RES_VERTICAL;

    gslayer[GS_LAYER_BACK_BUFFER].x      = 0;
    gslayer[GS_LAYER_BACK_BUFFER].y      = 0;
    gslayer[GS_LAYER_BACK_BUFFER].width  = GS_RES_HORIZONTAL;
    gslayer[GS_LAYER_BACK_BUFFER].height = GS_RES_VERTICAL;

    gs_cmd_clear();
    gs_cmd_flip();
    gs_cmd_syncbuf();
}

/*
 * get pointer to layer from enumeration
 *
 * RETURN
 *     pointer to layer structure or NULL on error
 */
struct gs_layer_t * gs_get_layer(int lnum)
{
    switch (lnum)
    {
        case GS_LAYER_FRONT_BUFFER:
            {
                gslayer[GS_LAYER_FRONT_BUFFER].mem = gs_active_buf;
                return &gslayer[GS_LAYER_FRONT_BUFFER];
            }
#if GS_DOUBLE_BUF
        case GS_LAYER_BACK_BUFFER:
            {
                gslayer[GS_LAYER_BACK_BUFFER].mem = gs_inactive_buf;
                return &gslayer[GS_LAYER_BACK_BUFFER];
            }
#endif
        default:
            debug_emsgf("unknown layer", "1xn", lnum);
    }

    return NULL;
}


/***********************************************
 * LCD command interface
 **********************************************/

/************* fill display with background ***************/
void gs_cmd_clear()
{
    union gs_pixel_t *mem;
    int h, w;

    mem = gs_inactive_buf;
    h   = GS_RES_VERTICAL;

    while (h--)
    {
        w = GS_RES_HORIZONTAL;
        while (w--)
        {
            (mem++)->value = gs.bgcolor.value;
        }
    }
}

#if GS_DOUBLE_BUF
/**************** switch back buffer and active buffer ******************/
void gs_cmd_flip()
{
    gs_hw_wait_vsync();
    gs.abuf ^= 0x01;
    gs_hw_set_active_buffer(gs_active_buf);
}

/*************** copy active buffer to back buffer ********************/
void gs_cmd_syncbuf()
{
    union gs_pixel_t *src;
    union gs_pixel_t *dst;
    int n;

    src = gs_active_buf;
    dst = gs_inactive_buf;
    n   = GS_PIXELNUM;

    while (n--)
        (dst++)->value = (src++)->value;
}
#endif /* GS_DOUBLE_BUF */

/************************ draw line *************************/
/*  ARGS
 *     layer     layer to draw to
 *     x0        x coordinate of start
 *     y0        y coordinate of start
 *     x1        x coordinate of end
 *     y1        y coordinate of end
 *     color     color of line
 */

void gs_cmd_line(struct gs_layer_t *layer, int x0, int y0, int x1, int y1, GS_COLOR color)
{

    int lengthX, lengthY, length, dx, dy;  
    union gs_pixel_t *mem;

    if (!layer)
    {
        debug_emsg("NULL layer");
        return;
    }

    if (x0 < 0 || y0 < 0 || x1 < 0 || y1 < 0 ||            
            x1 >= layer->width || y1 >= layer->height ||
            x0 >= layer->width || y0 >= layer->height)             /* sanity check */
        return;
    if (x0 > x1 && y0 > y1)    /* XY reorder */
    {
        int temp = x0;
        x0 = x1;
        x1 = temp;
        temp = y0;
        y0 = y1;
        y1 = temp;
    } 
    dx = (x1 - x0 >= 0 ? 1 : -1);
    dy = (y1 - y0 >= 0 ? 1 : -1);     
    lengthX = (x1 > x0) ? (x1 - x0) : (x0 - x1);  
    lengthY = (y1 > y0) ? (y1 - y0) : (y0 - y1);   
    length = (lengthX > lengthY) ? lengthX : lengthY;       
    if (length == 0)
    {
        mem = &layer->mem[layer->width * y0 + x0];    
        mem->value = color;
    }
    if (lengthY <= lengthX)
    {  
        int x = x0;    
        int y = y0;
        int d = -lengthX;
        length++;    
        while(length--)
        {
            mem = &layer->mem[layer->width * y + x];    
            mem->value = color;
            x += dx;
            d += 2 * lengthY;
            if (d > 0) {
                d -= 2 * lengthX;
                y += dy;
            }
        }
    }
    else
    {           
        int x = x0;    
        int y = y0;
        int d = - lengthY;
        length++;
        while(length--)
        {
            mem = &layer->mem[layer->width * y + x];    
            mem->value = color;
            y += dy;
            d += 2 * lengthX;
            if (d > 0) {
                d -= 2 * lengthY;
                x += dx;
            }
        }
    }
}

/********************* draw rectangle **********************/
/*  ARGS
 *     layer     layer to draw to
 *     x0        x coordinate of top left corner
 *     y0        y coordinate of top left corner
 *     x1        x coordinate of bottom right corner
 *     y1        y coordinate of bottom right corner
 *     lwidth   width of line
 *     color    color of line
 */
void gs_cmd_rect(struct gs_layer_t *layer, int x0, int y0, int x1, int y1, int lwidth, GS_COLOR color)
{
    int cnt;
    union gs_pixel_t *mem;

    if (!layer)
    {
        debug_emsg("NULL layer");
        return;
    }   

    if (lwidth == 0 || x0 < 0 || y0 < 0 || x1 < 0 || y1 < 0)    /* sanity check */
        return;

    if (x0 > x1)    /* X reorder */
    {
        cnt = x0;
        x0 = x1;
        x1 = cnt;
    }

    if (y0 > y1)    /* Y reorder */
    {
        cnt = y0;
        y0 = y1;
        y1 = cnt;
    }

    if (x1 >= layer->width || y1 >= layer->height)
        return;

    mem = &layer->mem[layer->width * y0 + x0];  /* upper and bottom lines */
    cnt = x1 - x0 + 1;
    while (cnt--)
    {
        mem->value = color;
        mem++;
    }
    mem = &layer->mem[layer->width * y1 + x0];
    cnt = x1 - x0 + 1;
    while (cnt--)
    {
        mem->value = color;
        mem++;
    }

    mem = &layer->mem[layer->width * y0 + x0];  /* left and right lines */
    cnt = y1 - y0 + 1;
    while (cnt--)
    {
        mem->value = color;
        mem += layer->width;
    }
    mem = &layer->mem[layer->width * y0 + x1];
    cnt = y1 - y0 + 1;
    while (cnt--)
    {
        mem->value = color;
        mem += layer->width;
    }

    /* XXX be care, recurse */
    if (lwidth > 1)
    {
        x0++; y0++;
        x1--; y1--;
        gs_cmd_rect(layer, x0, y0, x1, y1, lwidth - 1, color);
    }
}


/********************* draw filled rectangle **********************/
/*  ARGS
 *     layer     layer to draw to
 *     x0        x coordinate of top left corner
 *     y0        y coordinate of top left corner
 *     x1        x coordinate of bottom right corner
 *     y1        y coordinate of bottom right corner
 *     dir       gradient direction -   0=horizontal,    1=horiz_>|<,    2=vertical,    3=vert_>\<
 *     color    first color of line
 *     color1  second color of line
 */
void gs_cmd_filled_rect(struct gs_layer_t *layer, int x0, int y0, int x1, int y1, int dir, GS_COLOR color, GS_COLOR color1)
{
    int cnt;
    union gs_pixel_t *mem;

    if (!layer)
    {
        debug_emsg("NULL layer");
        return;
    }

    if (x0 < 0 || y0 < 0 || x1 < 0 || y1 < 0)    /* sanity check */
        return;
    if (x0 > x1)    /* X reorder */
    {
        cnt = x0;
        x0 = x1;
        x1 = cnt;
    }
    if (y0 > y1)    /* Y reorder */
    {
        cnt = y0;
        y0 = y1;
        y1 = cnt;
    }
    if (x1 >= layer->width || y1 >= layer->height)    /* sanity postcheck */
        return;

    if (color == color1)    /*  monocolor */
    {
        for (; y0 <= y1; y0++)
        {
            mem = &layer->mem[layer->width * y0 + x0];
            cnt = x1 - x0 + 1;
            while (cnt--)
                (mem++)->value = color;
        }
    }
    else
    {
        union gs_pixel_t pix;
        union gs_pixel_t pix1;
        pix.value = color;
        pix1.value = color1;
        int rwidth, rheight;
        rwidth = x1 - x0;
        rheight = y1 - y0;    
        float dred, dgreen, dblue,tred, tgreen, tblue;
        dred = pix1.red - pix.red;
        dgreen = pix1.green - pix.green;
        dblue = pix1.blue - pix.blue;

        if (dir == 0)    /* horizontal gradient*/
        {
            dred /= rwidth;
            dgreen /= rwidth;
            dblue /= rwidth;
            pix.value = color;
            tred = pix.red;
            tgreen = pix.green;
            tblue = pix.blue;
            for (; x0 <= x1; x0++)
            {
                mem = &layer->mem[layer->width * y0 + x0];               
                cnt = rheight + 1;
                while (cnt--)
                {                    
                    mem->value = pix.value;
                    mem += layer->width;
                } 
                tred +=dred;
                tgreen += dgreen;
                tblue += dblue;  
                pix.red = tred;
                pix.green = tgreen;
                pix.blue = tblue;
            } 
        }

        if (dir == 1)    /* horizontal  >|<  gradient*/
        {
            dred /= rwidth;
            dgreen /= rwidth;
            dblue /= rwidth;
            dred += dred;
            dgreen += dgreen;
            dblue += dblue;
            pix.value = color;
            tred = pix.red;
            tgreen = pix.green;
            tblue = pix.blue;

            for (; x0 <= (x1/2); x0++)
            {
                mem = &layer->mem[layer->width * y0 + x0];               
                cnt = rheight + 1;
                while (cnt--)
                {                    
                    mem->value = pix.value;
                    mem += layer->width;
                } 
                tred +=dred;
                tgreen += dgreen;
                tblue += dblue;  
                pix.red = tred + 0.5;
                pix.green = tgreen + 0.5;
                pix.blue = tblue + 0.5;
            }
            x0 = x1/2 + 1;
            for (; x0 <= x1; x0++)
            {
                mem = &layer->mem[layer->width * y0 + x0];               
                cnt = rheight + 1;
                while (cnt--)
                {                    
                    mem->value = pix.value;
                    mem += layer->width;
                } 
                tred -=dred;
                tgreen -= dgreen;
                tblue -= dblue;  
                pix.red = tred;
                pix.green = tgreen;
                pix.blue = tblue;
            }
        }

        if (dir == 2)      /* vertical gradient*/
        {
            dred /= rheight;
            dgreen /= rheight;
            dblue /= rheight;     
            pix.value = color;
            tred = pix.red;
            tgreen = pix.green;
            tblue = pix.blue;
            for (; y0 <= y1; y0++)
            {
                mem = &layer->mem[layer->width * y0 + x0];
                cnt = rwidth + 1;
                while (cnt--)
                {                    
                    mem->value = pix.value;
                    mem ++;
                }
                tred +=dred;
                tgreen += dgreen;
                tblue += dblue; 
                pix.red =  tred;
                pix.green = tgreen;
                pix.blue = tblue;            
            }
        }  
    }
}


/************************* draw ellipse ***********************/
/* ARGS
 *     layer    layer to draw to
 *     xc        x coordinate of center
 *     yc        y coordinate of center
 *     rh        horizontal radius
 *     rv        vertical radius
 *     ewidth width of ellipse
 *     color    color of line
 */
void gs_cmd_ellipse(struct gs_layer_t *layer, int xc, int yc, int rh, int rv, int ewidth, GS_COLOR color)
{
    int x, y; 
    int a, b; 
    int a2, ta2; 
    int b2, tb2; 
    int d, dx, dy;
    union gs_pixel_t *mem;

    if (!layer)
    {
        debug_emsg("NULL layer");
        return;
    }

    if (ewidth == 0 || xc < 0 || yc < 0 || rh < 0 || rv < 0)            
        return;
    if (xc <= rh || yc <= rv || ewidth >= rh || ewidth >= rv ||
            xc + rh >= layer->width || yc + rv >= layer->height)      /* sanity check */
        return;

    x = 0; 
    y = rh; 
    a = rv; 
    b = rh; 
    a2 = a * a; 
    ta2 = 2 * a2; 
    b2 = b * b; 
    tb2 = 2 * b2; 
    d = b2 - a2 * b + a2 / 4; 
    dx = 0; 
    dy = ta2 * b; 

    while (dx < dy) 
    { 
        mem = &layer->mem[layer->width * (yc + y) + xc + x];    
        mem->value = color;
        mem = mem - x - x; 
        mem->value = color; 
        mem = &layer->mem[layer->width * (yc - y) + xc + x];   
        mem->value = color;
        mem = mem - x - x; 
        mem->value = color;
        if (d > 0) 
        { 
            y--; 
            dy -= ta2; 
            d -= dy; 
        } 
        x++; 
        dx += tb2; 
        d += b2 + dx; 
    } 
    d += (3 * (a2 - b2) / 2 - (dx + dy)) / 2; 
    while (y >= 0) 
    { 
        mem = &layer->mem[layer->width * (yc + y) + xc + x];    
        mem->value = color;
        mem = mem - x - x;   
        mem->value = color; 
        mem = &layer->mem[layer->width * (yc - y) + xc + x];   
        mem->value = color;
        mem = mem - x - x;  
        mem->value = color;
        if (d < 0) 
        { 
            x++; 
            dx += tb2; 
            d += dx; 
        } 
        y--; 
        dy -= ta2; 
        d += a2 - dy; 
    } 

    if (ewidth > 1)                                                     /* XXX be care, recurse */
    {
        rh--; rv--;
        gs_cmd_ellipse(layer, xc, yc, rh, rv, ewidth - 1, color);
    }
}

/************************* draw filled ellipse ***********************/
/* ARGS
 *     layer    layer to draw to
 *     xc        x coordinate of center
 *     yc        y coordinate of center
 *     rh        horizontal radius
 *     rv        vertical radius
 *     color    color of ellipse
 */
void gs_cmd_fill_ellipse(struct gs_layer_t *layer, int xc, int yc, int rh, int rv, GS_COLOR color)
{
    int y, x, x0, x1, dx;
    int hh, ww, hhww;
    union gs_pixel_t *mem;

    if (!layer)
    {
        debug_emsg("NULL layer");
        return;
    }

    if (xc <= rh || yc <= rv || xc + rh >= layer->width || yc + rv >= layer->height)      /* sanity check */
        return;

    hh = rv * rv;
    ww = rh * rh;
    hhww = hh * ww;
    x0 = rh;
    dx = 0;
    for (x = -rh; x <= rh; x++)    // do the horizontal diameter
    {
        mem = &layer->mem[layer->width * yc + xc + x];    
        mem->value = color;   
    }
    for (y = 1; y <= rv; y++)    // now do both halves at the same time, away from the diameter
    {
        x1 = x0 - (dx - 1);     // try slopes of dx - 1 or more
        for ( ; x1 > 0; x1--)
        {
            if (x1*x1*hh + y*y*ww <= hhww)
                break;
            dx = x0 - x1;     // current approximation of the slope
            x0 = x1;
        }
        for (x = -x0; x <= x0; x++)
        {
            mem = &layer->mem[layer->width * (yc - y) + xc + x];    
            mem->value = color;
            mem = &layer->mem[layer->width * (yc + y) + xc + x];    
            mem->value = color;
        }
    }
}

