/* 
 *     This file is part of K14, the LCD controller.
 * 
 * Copyright (C) 2014 Dmitry Kobylin
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _GS_H_
#define _GS_H_

#include <types.h>

/**************************************************/
#define GS_RES_HORIZONTAL     800
#define GS_RES_VERTICAL       600
#define GS_PIXEL_DEPTH        2     /* 24 bpp */
#define GS_DOUBLE_BUF         1
/**************************************************/

#if (GS_PIXEL_DEPTH == 4)
    typedef uint32 GS_COLOR;

    #define GS_RGB2COLOR(r, g, b)     ((0xff << 24) | ((b) << 16) | ((g) << 8) | (r))
#elif (GS_PIXEL_DEPTH == 2)
    typedef uint16 GS_COLOR;
    #define GS_RGB2COLOR(r, g, b)     (((b) << 11) | ((g) << 5) | (r))

    #define GS_COLOR_WHITE       GS_RGB2COLOR(31, 63, 31)
    #define GS_COLOR_YELLOW      GS_RGB2COLOR(31, 63,  0)
    #define GS_COLOR_RED         GS_RGB2COLOR(31,  0,  0)
    #define GS_COLOR_GREEN       GS_RGB2COLOR( 0, 63,  0)
    #define GS_COLOR_BLUE        GS_RGB2COLOR( 0,  0, 31)
#else
    #error Unknown pixel size
#endif

#define GS_DEFAULT_COLOR      GS_RGB2COLOR(4, 8, 4)

#if GS_DOUBLE_BUF
    #define GS_MEM_BUFNUM         2     /* 2 for double buffered memory, 1 for single buffer */
#else
    #define GS_MEM_BUFNUM         1     /* 2 for double buffered memory, 1 for single buffer */
#endif

#define GS_PIXELNUM (GS_RES_HORIZONTAL * GS_RES_VERTICAL)
#define GS_MEM_SIZE (GS_PIXELNUM * GS_PIXEL_DEPTH)
/*
 * NOTE
 * Base address should be doubleword aligned. First buffer is aligned by linker script.
 * Check that second buffer also alligned. If not then add additional pixels.
 */
#define GS_BUF_SIZE (GS_PIXELNUM + (8 - (GS_PIXELNUM % 8)))

#pragma pack(push, 1)
union gs_pixel_t {
    GS_COLOR value;
#if (GS_PIXEL_DEPTH == 4)
    struct {
        uint8 red;
        uint8 green;
        uint8 blue;
        uint8 a; /* NOTE unused */
    };
#elif (GS_PIXEL_DEPTH == 2)
    struct {
        uint8 red   : 5;
        uint8 green : 6;
        uint8 blue  : 5;
    };
#endif
};
#pragma pack(pop)

struct gs_t {
#if GS_DOUBLE_BUF
    int abuf;        /* current active buffer (displayed on LCD) */
#endif
    union gs_pixel_t bgcolor;  /* background color */
};

#if GS_DOUBLE_BUF
    #define gs_active_buf   (gs_mem[(gs.abuf & 0x01)])
    #define gs_inactive_buf (gs_mem[(gs.abuf & 0x01) ^ 0x01])
#else
    #define gs_active_buf   (gs_mem[0])
    #define gs_inactive_buf  gs_active_buf
#endif

struct gs_layer_t {
    int x;
    int y;
    int width;
    int height;

    union gs_pixel_t *mem;
};

#define GS_LAYER_FRONT_BUFFER   0
#if GS_DOUBLE_BUF
#define GS_LAYER_BACK_BUFFER    1
#endif

extern struct gs_t gs;
extern union gs_pixel_t gs_mem[GS_MEM_BUFNUM][GS_BUF_SIZE];

#define GS_BUFFER_SIZE    4096
extern uint8 gs_buffer[GS_BUFFER_SIZE];

void gs_init();
struct gs_layer_t *gs_get_layer(int lnum);

/***********************************************
 * LCD command interface
 **********************************************/
void gs_cmd_clear();
#if GS_DOUBLE_BUF
    void gs_cmd_flip();
    void gs_cmd_syncbuf();
#else
    #define gs_cmd_flip()
    #define gs_cmd_syncbuf()
#endif
void gs_cmd_line(struct gs_layer_t *layer, int x0, int y0, int x1, int y1, GS_COLOR color);
void gs_cmd_rect(struct gs_layer_t *layer, int x0, int y0, int x1, int y1, int lwidth, GS_COLOR color);
void gs_cmd_filled_rect(struct gs_layer_t *layer, int x0, int y0, int x1, int y1, int dir, GS_COLOR color, GS_COLOR color1);
void gs_cmd_ellipse(struct gs_layer_t *layer, int xc, int yc, int rh, int rv, int ewidth, GS_COLOR color);
void gs_cmd_fill_ellipse(struct gs_layer_t *layer, int xc, int yc, int rh, int rv, GS_COLOR color);
/***********************************************
 * 
 **********************************************/

#include "gs_text.h"
#include "gs_data.h"
#include "gs_image.h"
#include "gs_image2.h"

#endif /* !_GS_H_ */

