/* 
 *     This file is part of K14, the LCD controller.
 * 
 * Copyright (C) 2014 Dmitry Kobylin
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <LPC177x_8x.h>
#include <LPC177x_8x_bits.h>
#include <gpio.h>
#include <debug.h>
#include <stimer.h>
#include <string.h>
#include <clk_cfg.h>
#include "gs.h"
#include "gs_hw.h"
#include "../irqp.h"

#define DEBUG_GS_HW

#ifdef DEBUG_GS_HW
    #define DPRINT(fmt, ...) dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif


#define LCD_CLKRATE   (48 * 1000 * 1000) /* Hz */

#define LCD_CLKDIV    (CLK_CCLK / LCD_CLKRATE)
#if (LCD_CLKDIV > 31)
    #error LCD clock divider too high
#endif
#define LCD_HSW       30 /* clocks, horizontal synchonization pulse width */
#define LCD_HFP       30 /* clocks, horizontal front porch */
#define LCD_HBP       60 /* clocks, horizontal back porch */
#define LCD_VSW       25 /* lines, vertical synchronization pulse width */
#define LCD_VFP        2 /* lines,  vertical front porch  */
#define LCD_VBP        8 /* lines,  vertical back porch */


static const struct gpio_t lcd_reset = {LPC_GPIO0, (1 << 15)};
#define GSP_HW_RESET_SET()         {gpio_drive(&lcd_reset, 0);}
#define GSP_HW_RESET_RELEASE()     {gpio_drive(&lcd_reset, 1);}

/*
 *
 */
void gs_hw_init()
{
    /*
     * configure pins
     *
     * P2[0]    LCD_PWR    (FUNC 7) -> NC ?
     * P2[1]    LCD_LE     (FUNC 7) -> NC   XXX backlight control?
     * P2[2]    LCD_DCLK   (FUNC 7) -> PCLK
     * P2[3]    LCD_FP     (FUNC 7) -> VS
     * P2[4]    LCD_ENAB_M (FUNC 7) -> DE
     * P2[5]    LCD_LP     (FUNC 7) -> HS
     *
     * P0[4]    LCD_VD0    (FUNC 7) -> R0
     * P0[5]    LCD_VD1    (FUNC 7) -> R1
     * P4[28]   LCD_VD2    (FUNC 7) -> R2
     * P4[29]   LCD_VD3    (FUNC 7) -> R3
     * P2[6]    LCD_VD4    (FUNC 7) -> R4
     * P2[7]    LCD_VD5    (FUNC 7) -> R5
     * P2[8]    LCD_VD6    (FUNC 7) -> R6
     * P2[9]    LCD_VD7    (FUNC 7) -> R7
     *
     * P0[6]    LCD_VD8    (FUNC 7) -> G0
     * P0[7]    LCD_VD9    (FUNC 7) -> G1
     * P1[20]   LCD_VD10   (FUNC 7) -> G2
     * P1[21]   LCD_VD11   (FUNC 7) -> G3
     * P1[22]   LCD_VD12   (FUNC 7) -> G4
     * P1[23]   LCD_VD13   (FUNC 7) -> G5
     * P1[24]   LCD_VD14   (FUNC 7) -> G6
     * P1[25]   LCD_VD15   (FUNC 7) -> G7
     *
     * P0[8]    LCD_VD16   (FUNC 7) -> B0
     * P0[9]    LCD_VD17   (FUNC 7) -> B1
     * P2[12]   LCD_VD18   (FUNC 7) -> B2
     * P2[13]   LCD_VD19   (FUNC 7) -> B3
     * P1[26]   LCD_VD20   (FUNC 7) -> B4
     * P1[27]   LCD_VD21   (FUNC 7) -> B5
     * P1[28]   LCD_VD22   (FUNC 7) -> B6
     * P1[29]   LCD_VD23   (FUNC 7) -> B7
     *
     * P0[15]                       -> Reset
     */
    gpio_setdir(&lcd_reset, GPIO_DIR_OUTPUT);
    GSP_HW_RESET_SET();

    LPC_IOCON_FUNC_SET(P2_2,  7); /*   LCD_DCLK    -> PCLK */
    LPC_IOCON_FUNC_SET(P2_3,  7); /*   LCD_FP      -> VS */
    LPC_IOCON_FUNC_SET(P2_4,  7); /*   LCD_ENAB_M  -> DE */
    LPC_IOCON_FUNC_SET(P2_5,  7); /*   LCD_LP      -> HS */

    LPC_IOCON_FUNC_SET(P0_4,  7); /*   LCD_VD0    -> R0 */
    LPC_IOCON_FUNC_SET(P0_5,  7); /*   LCD_VD1    -> R1 */
    LPC_IOCON_FUNC_SET(P4_28, 7); /*   LCD_VD2    -> R2 */
    LPC_IOCON_FUNC_SET(P4_29, 7); /*   LCD_VD3    -> R3 */
    LPC_IOCON_FUNC_SET(P2_6,  7); /*   LCD_VD4    -> R4 */
    LPC_IOCON_FUNC_SET(P2_7,  7); /*   LCD_VD5    -> R5 */
    LPC_IOCON_FUNC_SET(P2_8,  7); /*   LCD_VD6    -> R6 */
    LPC_IOCON_FUNC_SET(P2_9,  7); /*   LCD_VD7    -> R7 */

    LPC_IOCON_FUNC_SET(P0_6,  7); /*   LCD_VD8    -> G0 */
    LPC_IOCON_FUNC_SET(P0_7,  7); /*   LCD_VD9    -> G1 */
    LPC_IOCON_FUNC_SET(P1_20, 7); /*   LCD_VD10   -> G2 */
    LPC_IOCON_FUNC_SET(P1_21, 7); /*   LCD_VD11   -> G3 */
    LPC_IOCON_FUNC_SET(P1_22, 7); /*   LCD_VD12   -> G4 */
    LPC_IOCON_FUNC_SET(P1_23, 7); /*   LCD_VD13   -> G5 */
    LPC_IOCON_FUNC_SET(P1_24, 7); /*   LCD_VD14   -> G6 */
    LPC_IOCON_FUNC_SET(P1_25, 7); /*   LCD_VD15   -> G7 */

    LPC_IOCON_FUNC_SET(P0_8,  7); /*   LCD_VD16   -> B0 */
    LPC_IOCON_FUNC_SET(P0_9,  7); /*   LCD_VD17   -> B1 */
    LPC_IOCON_FUNC_SET(P2_12, 7); /*   LCD_VD18   -> B2 */
    LPC_IOCON_FUNC_SET(P2_13, 7); /*   LCD_VD19   -> B3 */
    LPC_IOCON_FUNC_SET(P1_26, 7); /*   LCD_VD20   -> B4 */
    LPC_IOCON_FUNC_SET(P1_27, 7); /*   LCD_VD21   -> B5 */
    LPC_IOCON_FUNC_SET(P1_28, 7); /*   LCD_VD22   -> B6 */
    LPC_IOCON_FUNC_SET(P1_29, 7); /*   LCD_VD23   -> B7 */

    LPC_SC->PCONP |= PCONP_PCLCD;
    LPC_LCD->CTRL = 0; /* disable LCD controller */

    LPC_SC->LCD_CFG = (LCD_CLKDIV - 1);

#define LCD_POL_CLKSEL_CCLK        (0 << 5)  /* clock source is CCLK */
#define LCD_POL_CLKSEL_CLKIN       (1 << 5)  /* clock source is LCD_CLKIN */
#define LCD_POL_IVS_FP_ACTIVE_HIGH (0 << 11) /* frame pulse (vsync) active HIGH */
#define LCD_POL_IVS_FP_ACTIVE_LOW  (1 << 11) /* frame pulse (vsync) active LOW */
#define LCD_POL_IHS_LP_ACTIVE_HIGH (0 << 12) /* line pulse (hsync) active HIGH */
#define LCD_POL_IHS_LP_ACTIVE_LOW  (1 << 12) /* line pulse (hsync) active LOW */
#define LCD_POL_IPC_CHANGE_RISING  (0 << 13) /* data is driven on the LCD lines on the RISING edge */
#define LCD_POL_IPC_CHANGE_FALLING (1 << 13) /* data is driven on the LCD lines on the FALLING edge */
#define LCD_POL_IOE_ACTIVE_HIGH    (0 << 14) /* LCD_ENAB_M pin is active HIGH */
#define LCD_POL_IOE_ACTIVE_LOW     (1 << 14) /* LCD_ENAB_M pin is active HIGH  */
#define LCD_POL_BCD                (1 << 26) /* bypass pixel clock divider */
    LPC_LCD->POL = 
        LCD_POL_CLKSEL_CCLK |
        LCD_POL_IVS_FP_ACTIVE_LOW |
        LCD_POL_IHS_LP_ACTIVE_LOW |
        LCD_POL_BCD |
//        LCD_POL_IPC_CHANGE_FALLING |
        ((GS_RES_HORIZONTAL - 1) << 16); /* CPL, clock per line. */

    LPC_LCD->TIMH =
        ((GS_RES_HORIZONTAL / 16 - 1) << 2) | /* PPL, pixel per line */
        ((LCD_HSW - 1) << 8)  |               /* HSW, horizontal synchonization pulse width */
        ((LCD_HFP - 1) << 16) |               /* HFP, horizontal front porch */
        ((LCD_HBP - 1) << 24);                /* HBP, horizontal back porch */

    LPC_LCD->TIMV = 
        (GS_RES_VERTICAL - 1) | /* Lines per panel. This is the number of active lines per screen */
        ((LCD_VSW - 1) << 10) | /* VSW, vertical synchronization pulse width */
        ((LCD_VFP - 1) << 16) | /* VFP, vertical front porch */
        ((LCD_VBP - 1) << 24);  /* VBP, vertical back porch */

    gs_hw_set_active_buffer(gs_active_buf); /* NOTE gsp.mem should be initialized */
#define LCD_CTRL_LCDEN          (1 << 0)    /* LCD enable control bit */
#define LCD_CTRL_BPP_1BPP       (0 << 1)    /* LCD bits per pixel, 1 bpp */
#define LCD_CTRL_BPP_2BPP       (1 << 1)    /* 2 bpp */
#define LCD_CTRL_BPP_4BPP       (2 << 1)    /* 4 bpp */
#define LCD_CTRL_BPP_8BPP       (3 << 1)    /* 8 bpp */
#define LCD_CTRL_BPP_16BPP      (4 << 1)    /* 16 bpp */
#define LCD_CTRL_BPP_24BPP      (5 << 1)    /* 24 bpp (TFT panel only) */
#define LCD_CTRL_BPP_16BPP_565  (6 << 1)    /* 16 bpp, 5:6:5 mode */
#define LCD_CTRL_BPP_12BPP_444  (7 << 1)    /* 12 bpp, 4:4:4 mode */
#define LCD_CTRL_LCDBW_COLOR    (0 << 4)    /* STN LCD is color */
#define LCD_CTRL_LCDBW_MONOCH   (1 << 4)    /* STN LCD is monochrome */
#define LCD_CTRL_LCDTFT_STN     (0 << 5)    /* LCD is STN. Use grayscaller */
#define LCD_CTRL_LCDTFT_TFT     (1 << 5)    /* LCD is TFT. Do not use grayscaller */
#define LCD_CTRL_LCDMONO8_4BIT  (0 << 6)    /* monochrome LCD uses a 4-bit interface */
#define LCD_CTRL_LCDMONO8_8BIT  (1 << 6)    /* monochrome LCD uses a 8-bit interface */
#define LCD_CTRL_LCDDUAL_SINGLE (0 << 7)    /* STN LCD interface is single-panel */
#define LCD_CTRL_LCDDUAL_DUAL   (1 << 7)    /* STN LCD interface is dual-panel */
#define LCD_CTRL_BGR_RGB        (0 << 8)    /* RGB: normal output */
#define LCD_CTRL_BGR_BGR        (1 << 8)    /* BGR: red and blue swapped */
#define LCD_CTRL_BEPO_BYTE_LE   (0 << 9)    /* little endian byte order in memory */
#define LCD_CTRL_BEPO_BYTE_BE   (1 << 9)    /* big endian byte order in memory */
#define LCD_CTRL_BEPO_PIXEL_LE  (0 << 10)   /* little endian ordering within byte */
#define LCD_CTRL_BEPO_PIXEL_BE  (1 << 10)   /* big endian ordering within byte */
#define LCD_CTRL_LCDPWR         (1 << 11)   /* LCD power enable */
#define LCD_CTRL_LCDVCOMP_VS    (0 << 12)   /* LCD Vertical Compare interrupt. Start of vertical sync. */
#define LCD_CTRL_LCDVCOMP_BP    (1 << 12)   /* start of back porch */
#define LCD_CTRL_LCDVCOMP_AV    (2 << 12)   /* start of active video  */
#define LCD_CTRL_LCDVCOMP_FP    (3 << 12)   /* start of front porch */
#define LCD_CTRL_WATERMARK_4    (0 << 16)   /* LCD DMA request is generated when either of the DMA FIFOs have four or more empty locations.  */
#define LCD_CTRL_WATERMARK_8    (1 << 16)   /* LCD DMA request is generated when either of the DMA FIFOs have eight or more empty locations. */
    LPC_LCD->CTRL =
        LCD_CTRL_BPP_16BPP_565 |
        LCD_CTRL_LCDTFT_TFT;
    LPC_LCD->CTRL |= LCD_CTRL_BGR_BGR;        
    LPC_LCD->CTRL |= LCD_CTRL_LCDEN; /* enable LCD controller */
    stimer_wait(10);
    GSP_HW_RESET_RELEASE();
    LPC_LCD->CTRL |= LCD_CTRL_LCDPWR; /* set LCD_POWER, LCD_VD[23:0] go to active state */
    stimer_wait(20);

#define LCD_INTMSK_LNBUI   (1 << 2) /* next base address update interrupt */
    LPC_LCD->INTMSK = LCD_INTMSK_LNBUI;
//    NVIC_SetPriorityRaw(LCD_IRQn, LCD_IRQP);
}

/*
 *
 */
void gs_hw_set_active_buffer(union gs_pixel_t *mem)
{
    LPC_LCD->UPBASE = ((uint32)mem) & (~0x0000007);
    LPC_LCD->LPBASE = ((uint32)mem) & (~0x0000007);
}

/*
 *
 */
void LCD_Handler(void)
{
//    if (LPC_LCD->INTSTAT & LCD_INTMSK_LNBUI)
//    {
//        LPC_LCD->INTCLR = LCD_INTMSK_LNBUI;
//        NVIC_DisableIRQ(LCD_IRQn);
//    }
}

/*
 *
 */
void gs_hw_wait_vsync()
{
    LPC_LCD->INTCLR = LCD_INTMSK_LNBUI;
//    NVIC_EnableIRQ(LCD_IRQn);

    /*
     * XXX polled
     */
    while (!(LPC_LCD->INTSTAT & LCD_INTMSK_LNBUI))
        ;
}

