#include <debug.h>
#include "gs.h"
#include "gs_image2.h"
#include "../image/tjpgd/tjpgd.h"
#include "../media.h"

/*
 * NOTE
 * JD_FORMAT    Output pixel format 0:RGB888 (3 BYTE/pix), 1:RGB565 (1 WORD/pix)
 */

#define DEBUG_IMAGE2

#ifdef DEBUG_IMAGE2
    #define DPRINT(fmt, ...)  dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif

struct jpegdev_t {
    struct gs_layer_t *layer;
    int x;
    int y;
    void *file;
};

static JDEC jdec;

static UINT _jpeg_in_func(JDEC* jdec, BYTE* buff, UINT ndata);
static UINT _jpeg_out_func(JDEC* jdec, void* bitmap, JRECT* rect);

static struct jpegdev_t jpegdev;

/*
 *
 */
void gs_cmd_put_jpeg_from_file(struct gs_layer_t *layer, char *path, int x, int y)
{
    int result;

    jpegdev.file = media_open_file(path);
    if (!jpegdev.file)
    {
        debug_emsgf("failed to open file", "sn", path);
        return;
    }

    jpegdev.layer = layer;
    jpegdev.x     = x;
    jpegdev.y     = y;

    /*
     * jd_prepare (
     *     JDEC* jdec,                      // Pointer to blank decompression object
     *     UINT(*infunc)(JDEC*,BYTE*,UINT), // Pointer to input function
     *     void* work,                      // Pointer to work area
     *     UINT sz_work,                    // Size of the work area
     *     void* device                     // Device identifier for the session
     * );
     */
    result = jd_prepare(&jdec, _jpeg_in_func, gs_buffer, GS_BUFFER_SIZE, &jpegdev);
    if (result != JDR_OK)
    {
        debug_emsgf("JPEG prepare, ret code", "4dn", result);
        goto error;
    }

    /*
     * JRESULT jd_decomp (
     *     JDEC* jdec,                         // Pointer to valid decompressor object
     *     UINT(*outfunc)(JDEC*,void*,JRECT*), // Pointer to output function
     *     BYTE scale                          // Scaling factor
     * );
     */
    result = jd_decomp(&jdec, _jpeg_out_func, 0);   /* Start to decompress with 1/1 scaling */
    if (result == JDR_OK)
        goto out;

error:
    debug_emsgf("JPEG decomp, ret code", "4dn", result);
out:
    media_close_file(jpegdev.file);
}

/*
 * This function is the data input interface of the TJpgDec module. The corresponding
 * decompression session can be identified by the pointer to the device identifier
 * jdec->device passed to the 5th argument of jd_prepare function.
 *
 * ARGS
 *     jdec    Pointer to the decompression object
 *     buff    Pointer to buffer to store the read data
 *     ndata   Number of bytes to read
 *
 * RETURN
 *     Returns number of bytes read/removed. When a zero is returned, the
 *     jd_prepare and jd_decomp function aborts with JDR_INP.
 */
static UINT _jpeg_in_func(JDEC* jdec, BYTE* buff, UINT ndata)
{
    UINT n;
    static struct jpegdev_t *dev;

    dev = (struct jpegdev_t*)jdec->device;   /* Device identifier for the session (5th argument of jd_prepare function) */

    if (buff) {
        n = media_read_file(dev->file, (char*)buff, ndata);
        return n;
    } else {
        media_seek(dev->file, ndata, SEEK_CUR);
        return ndata;
    }    

    return 0;
}

/*
 * This function is the data output interface of the TJpgDec module. The
 * corresponding decompression session can be identified by the pointer to
 * the device identifier jdec->device passed to the 5th argument of jd_prepare
 * function.
 *
 * The bitmap is sent to the frame buffer or display device in this function.
 * The first pixel in the bitmap is the left-top of the rectangular, the
 * second one is next right and last pixel is the bottom-right of the rectangular.
 *
 * The size of rectangular varies from 1x1 to 16x16 depends on clipping,
 * scaling and sampling factor of the image. If the rectangular is out
 * of the frame buffer, it should be clipped in this function.
 *
 * The pixel format is defined by JD_FORMAT parameter in the configuration
 * option. When it is configured to RGB888, the bitmap is a BYTE array holds
 * the RGB data in 3 bytes per pixel: RRRRRRRR, GGGGGGGG, BBBBBBBB, RRRRRRRR,
 * GGGGGGGG, BBBBBBBB, ... When it is configured to RGB565, the bitmap
 * is a WORD array holds the RGB data in 1 word per pixel: RRRRRGGGGGGBBBBB,
 * RRRRRGGGGGGBBBBB, RRRRRGGGGGGBBBBB, ...
 *
 * ARGS
 *
 *     jdec    Pointer to the decompression object
 *     bitmap  RGB bitmap to be output
 *     rect    Rectangular region to output
 *
 * RETURN
 *     Normally returns 1. It lets TJpgDec to continue the decompressing
 *     process. When a 0 is returned, the jd_decomp function aborts with JDR_INTR.
 *     This is useful to interrupt the decompression process.
 *
 *
 */
static UINT _jpeg_out_func(JDEC* jdec, void* bitmap, JRECT* rect)
{
    int rows, cols, bwidth, pix;
    static struct jpegdev_t *dev;
    union gs_pixel_t *dst;
    union gs_pixel_t *src;
    union gs_pixel_t *src_base;

    dev = (struct jpegdev_t*)jdec->device;   /* Device identifier for the session (5th argument of jd_prepare function) */


    rows = rect->bottom - rect->top  + 1;
    cols = rect->right  - rect->left + 1;

    bwidth = cols;

    /* clip aread to draw */
    if (rect->right + dev->x >= dev->layer->width)
        cols = dev->layer->width - (rect->left + dev->x);
    if (rect->bottom + dev->y >= dev->layer->height)
        rows = dev->layer->height - (rect->top + dev->y);

    /* rectangular does not appear on screen - return */
    if (cols <= 0 || rows <= 0)
        return 1; /* NOTE does not interrupt decompression */

    src_base = (union gs_pixel_t *)bitmap;
    dst      = &dev->layer->mem[dev->layer->width * (rect->top + dev->y) + rect->left + dev->x];

    while (rows--)
    {
        pix = cols;
        src = src_base;
        while (pix--)
        {
            dst->value = src->value;
            dst++;
            src++;

//            *dst++ = *src++;
//            *dst++ = *src++;
//            *dst++ = *src++;
//            *dst++ = 0xff;
        }
        src_base += bwidth;
        dst      += dev->layer->width - cols;
    }

    return 1;
}

