#include <debug.h>
#include "gs.h"
#include "gs_text.h"

#define DEBUG_GS_TEXT

#ifdef DEBUG_GS_TEXT
    #define DPRINT(fmt, ...) dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif

static int _utf8_to_ucs2(char *utf8, uint16 *ucs2);

/*
 * Draw text to screen
 */
void gs_text_put(struct gs_layer_t *layer, char *text,int x, int y, int maxwidth,
        GS_COLOR fgcolor, GS_COLOR bgcolor, struct gs_font_t *font)
{
#define UCS2_CHAR  0
#define UCS2_PAGE  1
    uint8 ucs2[2];
    int clen;
    int width;
    uint8 *fdata;
    int chsize, chwidth;
    uint32 chdata;
#define PIXBUF_SIZE   32
    union gs_pixel_t *mem;

    if (!layer)
    {
        debug_emsg("NULL layer");
        return;
    }

    /* TODO font sanity check; x, y sanity check */
    if (font->magic != GS_FONT_MAGIC)
    {
        debug_emsg("invalid font magic");
        return;
    }

    chsize = font->maxwidth * font->height + 1 ;  /* + 1 is for width */

    if (y + font->height > layer->height)
    {
        DEBUG_WMSG("text does not fit to layer by height");
        return;
    }

    if (maxwidth <= 0)
        width = 0 - layer->width;
    else
        width = 0;

    while (*text)
    {
        clen = _utf8_to_ucs2(text, (uint16*)ucs2);
        if (clen == 0)
            break;
        text += clen;

        fdata = &font->data[
                chsize * (256 * font->ipage[ucs2[UCS2_PAGE]] + ucs2[UCS2_CHAR])
            ];
        chwidth = *fdata++;

        if (chwidth > PIXBUF_SIZE)
        {
            DEBUG_WMSGF("character to wide", "4dn", chwidth);
            return;
        }

        width += chwidth;
        if (width > maxwidth)
            return;

        if (x + chwidth > layer->width)
        {
            DEBUG_WMSG("text does not fit to layer by width");
            return;
        }

        /* put char */
        {
            int n, line, mask;

            line = font->height;
            mem  = &layer->mem[y * layer->width + x];
            while (line--)
            {
                mask = 0x01 << (chwidth - 1);

                /* XXX width is limited to 32 pixels */
                chdata  = *((uint32*)fdata);
                n       = chwidth;
                while (n--)
                {
                    if (chdata & mask)
                        mem->value = fgcolor;
                    else
                        mem->value = bgcolor;
                    mem++;
                    mask >>= 1;
                }
                fdata += font->maxwidth;
                mem += layer->width - chwidth;
            }
        }
        x += chwidth;
    }
}

/*
 *
 *
 * A little guide to UTF-8
 * =======================
 *
 *
 * Bits of      First       Last         Bytes in
 * code point   code point  code point   sequence  Byte 1      Byte 2      Byte 3      Byte 4      Byte 5      Byte 6
 * 7            U+0000      U+007F       1         0xxxxxxx
 * 11           U+0080      U+07FF       2         110xxxxx    10xxxxxx
 * 16           U+0800      U+FFFF       3         1110xxxx    10xxxxxx    10xxxxxx
 * 21           U+10000     U+1FFFFF     4         11110xxx    10xxxxxx    10xxxxxx    10xxxxxx
 * 26           U+200000    U+3FFFFFF    5         111110xx    10xxxxxx    10xxxxxx    10xxxxxx    10xxxxxx
 * 31           U+4000000   U+7FFFFFFF   6         1111110x    10xxxxxx    10xxxxxx    10xxxxxx    10xxxxxx    10xxxxxx
 *
 *
 * Russian "М" (M)
 *
 * Unicode code point   0x041C
 * UTF8 representation  0x9CD0
 *
 * 0xD0  11010000
 *          -----  MSB
 *
 * 0x9C  10011100
 *         ------  LSB
 *
 *        
 * 0b10.0001.1100
 *    4    1    C
 *
 */


/*
 * Convert utf8 character to ucs2 (BMP).
 *
 * ARGS
 *     utf8    pointer to utf8 character
 *     ucs2    pointer where code of converted character will be placed
 *
 * RETURN
 *     Size of utf8 character in bytes or zero if error was occured or end
 *     of line was reached.
 */
static int _utf8_to_ucs2(char *utf8, uint16 *ucs2)
{
    if (!(*utf8))
        return 0;

    *ucs2 = 0x0000;

    if (!(*utf8 & 0x80))
    {
        *ucs2 |= *utf8;
        return 1;
    }

    if ((*utf8 & 0xE0) == 0xC0)
    {
        *ucs2  =                (*(utf8 + 0) & 0x1F);
        *ucs2  = (*ucs2 << 6) | (*(utf8 + 1) & 0x3F);
        return 2;
    } else if ((*utf8 & 0xF0) == 0xE0) {
        *ucs2  =                (*(utf8 + 0) & 0x0F);
        *ucs2  = (*ucs2 << 6) | (*(utf8 + 1) & 0x3F);
        *ucs2  = (*ucs2 << 6) | (*(utf8 + 2) & 0x3F);
        return 3;
    } else {
        debug_emsg("decode error");
        return 0;
    }
}

