#include <debug.h>
#include <ssp.h>
#include <LPC177x_8x.h>
#include <LPC177x_8x_bits.h>
#include <crc.h>
#include "sdcard.h"
#include "sdcard_def.h"

#define DEBUG_SDCARD

#ifdef DEBUG_SDCARD
    #define DPRINT(fmt, ...) dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif

#define SD_BLOCK_SIZE    512

#define SDCARD_SSP   SSP1

/* 
 * SD_LO_CLK should not exceed 400 kHz.
 * SD_HI_CLK should not exceed 25 MHz.
 */
#define SD_LO_CLK   SSP_SPEED_100KHZ
#define SD_HI_CLK   SSP_SPEED_12MHZ

#define SD_CS(value)        do { if (value) LPC_GPIO0->SET  = (1 << 10); else LPC_GPIO0->CLR = (1 << 10); } while (0)
#define SD_CS_DIR_OUTPUT()  do {            LPC_GPIO0->DIR |= (1 << 10);} while (0)
//#define SD_CD()             (LPC_GPIO1->FIOPIN & (1 << 25))

#define SD_CARDTYPE_UNUSABLE       0x00
#define SD_CARDTYPE_SD_V1X         0x01
#define SD_CARDTYPE_SD_V200_SC     0x02
#define SD_CARDTYPE_SD_V200_HC     0x03
#define SD_CARDTYPE_SD_V200_XC     0x04

static int cardtype;

enum sd_expect_t {
    SD_EXPECT_R1,
    SD_EXPECT_R1b,
    SD_EXPECT_R2,
    SD_EXPECT_R3,
    SD_EXPECT_R7,
};

#define CS_RELEASE    1
#define CS_HOLD       0

static int sdcard_cmd(uint8 cmd,
        union sd_argument_t *arg, union sd_response_t *resp, enum sd_expect_t expect, int cs_release);
static int sdcard_cmd0();
static int sdcard_cmd8();
static int sdcard_cmd16(int blocklen);
static int sdcard_cmd58();
static int sdcard_acmd41();
static int sdcard_cmd16(int blocklen);
static void _swap32(uint32 *v);
static void _swap16(uint16 *v);

/*
 * RETURN
 *     0 on success, -1 on error
 */
int sdcard_init()
{
    {
        SD_CS(1);
        SD_CS_DIR_OUTPUT();
        ssp_setup(SDCARD_SSP, SD_LO_CLK, SSP_MODE_CPOL0_CPHA0);
    }

    cardtype = SD_CARDTYPE_UNUSABLE;

//    /* 
//     * Transfer 74 or more clocks with CS and DI high - the card
//     * will enter it's native operation mode
//     */
//    SD_CS(1);
//    ssp_wr(SDCARD_SSP, NULL, NULL, 10);

    if (sdcard_cmd0() < 0)
        return -1;
    if (sdcard_cmd8() < 0)
        return -1;
    if (sdcard_cmd58() < 0)
        return -1;
    if (sdcard_acmd41() < 0)
        return -1;
    if (sdcard_cmd58() < 0)
        return -1;
    if (sdcard_cmd16(SD_BLOCK_SIZE) < 0)
        return -1;

    ssp_setup(SDCARD_SSP, SD_HI_CLK, SSP_MODE_CPOL0_CPHA0);

    DPRINT("sn", "SD init OK");

    return 0;
}

#define RESPONSE_TIMEOUTE_NCYCLE   20

/*
 *
 */
int sdcard_detect()
{
//    if (SD_CD())
//        return 0;
//
    return 1;
}

/*
 *
 * RETURN
 *     zero on success, -1 on error
 */
static int sdcard_cmd0()
{
    union sd_argument_t arg;
    union sd_response_t resp;
    int retry;

    retry = 5;

    arg.value = 0;
    while (retry--)
    {
        if (sdcard_cmd(SD_CMD0_GO_IDLE_STATE, &arg, &resp, SD_EXPECT_R1, CS_RELEASE) < 0)
        {
            debug_emsg("CMD0 failed");
            continue;
        }

        if (resp.r1.val != 0x01)
        {
            dprint("<CMD0 invalid response >1xn", resp.r1.val);
            continue;
        }

        return 0;
    }

    debug_emsg("retry limit reached");

    return -1;
}

/*
 * CMD8 command (SEND_IF_COND)
 *
 * RETURN
 *     0 if valid response was received, -1 otherwise
 */
static int sdcard_cmd8()
{
    union sd_argument_t arg;
    union sd_response_t resp;
    int retry;

#define CHECK_PATTERN    0xAB
    arg.value = 0;
    arg.cmd08.check_pattern    = 0xAB;
    arg.cmd08.voltage_accepted = VOLTAGE_ACCEPTED_27_36;

    retry = 5;
    while (retry--)
    {
        if (sdcard_cmd(SD_CMD8_SEND_IF_COND, &arg, &resp, SD_EXPECT_R7, CS_RELEASE) == 0)
        {
            if (resp.r1.val != 0x01)
            {
                debug_emsgf("R1 ", "1xn", resp.r1.val);
                continue;
            }

            /* 
             * NOTE
             * Check pattern not necessary. Card does not response if 
             * voltage not supported or sdcard is Ver1.X.
             */
            if (resp.r7.check_pattern == CHECK_PATTERN && resp.r7.voltage_accepted == VOLTAGE_ACCEPTED_27_36)
            {
                return 0;
            } else {
                debug_emsgf("CMD8 check pattern failed, voltage not accepted, or ver1.x: ",
                        "<R1 >1x< R7 >4xn", resp.r1.val, resp.val32);
                return -1;
            }
        }
    }

    debug_emsg("retry limit reached");
    return -1;
}

/*
 * CMD16 (SET_BLOCKLEN)
 *
 * RETURN
 *     0 if valid response was received, -1 otherwise
 */
static int sdcard_cmd16(int blocklen)
{
    int retry;
    union sd_argument_t arg;
    union sd_response_t resp;

    arg.value = blocklen;

    retry = 20;
    while (retry--)
    {
        if (sdcard_cmd(SD_CMD16_SET_BLOCKLEN, &arg, &resp, SD_EXPECT_R1, CS_RELEASE) == 0)
        {
            if (resp.r1.val & 0xFE)
            {
                debug_emsgf("R1 ", "1xn", resp.r1.val);
                continue;
            }

            return 0;
        }
    }

    debug_emsg("retry limit reached");
    return -1;
}

/*
 * CMD58 command (SEND_IF_COND)
 *
 * RETURN
 *     0 if valid response was received, -1 otherwise
 */
static int sdcard_cmd58()
{
    union sd_argument_t arg;
    union sd_response_t resp;
    int retry;

    arg.value = 0;

    retry = 5;
    while (retry--)
    {
        if (sdcard_cmd(SD_CMD58_READ_OCR, &arg, &resp, SD_EXPECT_R3, CS_RELEASE) == 0)
        {
            if (resp.r1.val & 0xFE)
            {
                debug_emsgf("R1 ", "1xn", resp.r1.val);
                continue;
            }

            DEBUG_IMSGF("CCS ", "1dn", resp.r3.ccs);
            if (resp.r3.ccs)
                cardtype = SD_CARDTYPE_SD_V200_SC;
            else
                cardtype = SD_CARDTYPE_SD_V200_HC;

            return 0;
        }
    }

    debug_emsg("retry limit reached");
    return -1;
}

/*
 * CMD55 (APP_CMD)
 *
 * RETURN
 *     0 if valid response was received, -1 otherwise
 */
static int sdcard_cmd55()
{
    int retry;
    union sd_argument_t arg;
    union sd_response_t resp;

    arg.value = 0;
//    arg.cmd55.rca = cardstate.rca;
    arg.cmd55.rca = 0;

    retry = 5;
    while (retry--)
    {
        if (sdcard_cmd(SD_CMD55_APP_CMD, &arg, &resp, SD_EXPECT_R1, CS_RELEASE) == 0)
        {
            return 0;
        }
    }
    debug_emsg("retry limit reached");
    return -1;
}



/*
 * ACMD41 (SD_SEND_OP_COND)
 *
 * RETURN
 *     0 if valid response was received, -1 otherwise
 */
static int sdcard_acmd41()
{
    int retry;
    union sd_argument_t arg;
    union sd_response_t resp;

    arg.value = 0;
    /* XXX */
    arg.acmd41.v27_28 = 0;
    arg.acmd41.v28_29 = 0;
    arg.acmd41.v29_30 = 0;
    arg.acmd41.v30_31 = 0;
    arg.acmd41.v31_32 = 1;
    arg.acmd41.v32_33 = 1;
    arg.acmd41.v33_34 = 0;
    arg.acmd41.v34_35 = 0;
    arg.acmd41.v35_36 = 0;
    arg.acmd41.hcs    = 1;

    retry = 1000;
    while (retry--)
    {
        if (sdcard_cmd55() < 0)
            continue;

        if (sdcard_cmd(SD_ACMD41_SD_SEND_OP_COND, &arg, &resp, SD_EXPECT_R1, CS_RELEASE) < 0)
            continue;

        if (resp.r1.in_idle_state)
        {
            DEBUG_IMSG("idle");
            continue;
        }

        DEBUG_IMSG("done");
        return 0;
    }

    debug_emsg("retry limit reached");
    return -1;
}


/*
 *
 * RETURN
 *     0 if valid response was received, -1 otherwise
 */
int sdcard_read_block(uint32 blocknum, uint8 *buf)
{
    union sd_argument_t arg;
    union sd_response_t resp;
    int retry;
    int timeout;
    uint8 token;
    int ret;
    uint16 crc0, crc1;

    ret = -1;
    arg.value = blocknum;

    retry = 10;
    do
    {
        if (sdcard_cmd(SD_CMD17_READ_SINGLE_BLOCK, &arg, &resp, SD_EXPECT_R1, CS_HOLD) == 0)
        {
            if (resp.r1.val != 0)
            {
                debug_emsgf("R1 ", "1xn", resp.r1.val);
                continue;
            }

            break;
        }
    } while (retry--);

    if (retry < 0)
    {
        debug_emsg("retry limit reached");
        goto out;
    }

    /*
     * Physical Layer Simplified Specification Version 4.10, page
     */
#define CMD17_DATA_TOKEN       0xFE

    timeout = 100000;
    while (timeout--)
    {
        /* lookup for data or error token */
        ssp_wr(SDCARD_SSP, NULL, &token, 1);
        if (token == 0xff)
            continue;

        if (token == CMD17_DATA_TOKEN)
        {
            /* read data block */
            ssp_wr(SDCARD_SSP, NULL, buf, SD_BLOCK_SIZE);

            /* read CRC */
            ssp_wr(SDCARD_SSP, NULL, (uint8*)&crc0, 2);
            _swap16(&crc0);

            crc1 = 0;
            crc16ccitt_update(&crc1, buf, SD_BLOCK_SIZE);

            if (crc0 != crc1)
            {
                debug_emsgf("CRC mismatch", "2x_2xn", crc0, crc1);
                goto out;
            }

            ret = 0;
            goto out;
        }
        if ((token & 0xF0) == 0)
        {
            debug_emsgf("error token received", "1xn", token);
            ret = -1;
            goto out;
        }
    }
    
    debug_emsg("timeout");

out:
    SD_CS(1);
    return ret;
}

/*
 *
 */
static void _swap32(uint32 *v)
{
    uint32 value;

    value  = (*v & 0xff000000) >> 24;
    value |= (*v & 0x00ff0000) >> 8;
    value |= (*v & 0x0000ff00) << 8;
    value |= (*v & 0x000000ff) << 24;

    *v = value;
}

/*
 *
 */
static void _swap16(uint16 *v)
{
    uint16 value;

    value  = (*v & 0xff00) >> 8;
    value |= (*v & 0x00ff) << 8;

    *v = value;
}



/*
 *
 * RETURN
 *     zero on success, -1 on error
 */
static int sdcard_cmd(uint8 cmd,
        union sd_argument_t *arg, union sd_response_t *resp, enum sd_expect_t expect, int cs_release)
{
    uint8 crc;
    int ncycle;
    int ret;

    ret = 0;

    cmd |= 0x40;
    _swap32(&arg->value);

    /*
     * NOTE
     * When once the card enters SPI mode, the CRC feature is disabled
     * and the CRC is not checked by the card,
     * so that command transmission routine can be written with the
     * hardcorded CRC value that valid for only CMD0 and CMD8 with the argument of zero.
     * The CRC feature can also be switched with CMD59.
     */

    crc = 0;
    if (cmd == SD_CMD0_GO_IDLE_STATE || SD_CMD8_SEND_IF_COND)
    {
        crc7_update(&crc, &cmd, 1);
        crc7_update(&crc, (uint8*)&arg->value, 4);
    }

    crc <<= 1;
    crc |= 1;

    SD_CS(0);
    {
        ssp_wr(SDCARD_SSP, &cmd, NULL, 1);
        ssp_wr(SDCARD_SSP, (uint8*)&arg->value, NULL, 4);
        ssp_wr(SDCARD_SSP, &crc, NULL, 1);

        /* wait for response */
        ncycle = RESPONSE_TIMEOUTE_NCYCLE;
        do
        {
            ssp_wr(SDCARD_SSP, NULL, &resp->val[0], 1);
            if (resp->r1.zero == 0)
                break;
        } while (ncycle--);

        if (ncycle < 0)
        {
            debug_emsg("wait response timeout");
            ret = -1;
            goto out;
        }

        switch (expect)
        {
            case SD_EXPECT_R1:
                goto out;
            case SD_EXPECT_R3:
            case SD_EXPECT_R7:
                ssp_wr(SDCARD_SSP, NULL, &resp->val[1], 4);
                _swap32(&resp->val32);
                break;
            default:
                debug_emsg("unknown expect");
                ret = -1;
                goto out;
        }
    }
out:
    if (cs_release)
        SD_CS(1);
    return ret;
}

