#include <LPC177x_8x_bits.h>
#include <debug.h>
#include "ssp.h"

#define DEBUGSSP

#ifdef DEBUGSSP
    #define DPRINT(fmt, ...) dprint(fmt, __VA_ARGS__)
#else
    #define DPRINT(fmt, ...)
#endif


#define SSP_CR1_SSE (1 << 1)

#define SSP_SR_TFE  (1 << 0)
#define SSP_SR_RNE  (1 << 2)
#define SSP_SR_RFF  (1 << 3)
#define SSP_SR_BSY  (1 << 4)

struct ssp_t ssp[SSP_COUNT];

/*
 *
 */
void ssp_setup(struct ssp_t *sspif, uint32 speed, uint32 mode)
{
    if (sspif == SSP0)
    {
        sspif->pm = LPC_SSP0;

        LPC_SC->PCONP |= PCONP_PCSSP0;

        /*
         * P1.20/SCK
         * P1.23/MISO
         * P1.24/MOSI
         */
        LPC_IOCON_FUNC_SET(P1_20, LPC_IOCON_P1_20_SSP0_SCK);
        LPC_IOCON_FUNC_SET(P1_23, LPC_IOCON_P1_23_SSP0_MISO);
        LPC_IOCON_FUNC_SET(P1_24, LPC_IOCON_P1_24_SSP0_MOSI);
    } else if (sspif == SSP1) {
        sspif->pm = LPC_SSP1;

        LPC_SC->PCONP |= PCONP_PCSSP1;

        /*
         * P0.7/SCK
         * P0.8/MISO
         * P0.9/MOSI
         */
        LPC_IOCON_FUNC_SET(P0_7, LPC_IOCON_P0_7_SSP1_SCK);
        LPC_IOCON_FUNC_SET(P0_8, LPC_IOCON_P0_8_SSP1_MISO);
        LPC_IOCON_FUNC_SET(P0_9, LPC_IOCON_P0_9_SSP1_MOSI);
    } else {
        /* NOTREACHED */
        DEBUG_EMSG("unknown SSP interface");
        return;
    }

    sspif->pm->CR1 &= ~SSP_CR1_SSE;

    /*
     * 8-bit transfer, SPI frame format
     */
    sspif->pm->CR0 = 0x07;

    ssp_setspeed(sspif, speed);
    ssp_setmode(sspif, mode);

    /* enable SSP NOTE not necessary (look at setmode function) */
    sspif->pm->CR1 |= SSP_CR1_SSE;

    /* read out all bytes from receive FIFO */
    while ((sspif->pm->SR & SSP_SR_RNE) && !(sspif->pm->SR & SSP_SR_TFE))
        sspif->pm->DR;
}

/*
 *
 */
void ssp_setmode(struct ssp_t *sspif, uint32 mode)
{
    /* disable SSP */
    sspif->pm->CR1 &= ~SSP_CR1_SSE;

    sspif->pm->CR0 &= ~(0x03 << 6);
    sspif->pm->CR0 |=  (mode << 6);

    /* enable SSP */
    sspif->pm->CR1 |= SSP_CR1_SSE;

    /* save mode value */
    sspif->mode = mode;
}

/*
 *
 */
void ssp_setspeed(struct ssp_t *sspif, uint32 speed)
{
    /* disable SSP */
    sspif->pm->CR1 &= ~SSP_CR1_SSE;

    sspif->pm->CPSR &= ~0xff;
    sspif->pm->CPSR |= speed & 0x00ff;

    sspif->pm->CR0 &= ~(0xff << 8);
    sspif->pm->CR0 |= speed & 0xff00;

    /* enable SSP */
    sspif->pm->CR1 |= SSP_CR1_SSE;

    /* save speed value */
    sspif->speed = speed;
}

/*
 * read/write SPI frame
 *
 * ARGS
 *     sspif   pointer to SSP interface
 *     tx      pointer to transfer buffer, NULL if should not care
 *     rx      pointer to receive buffer, NULL if should not care
 *     len     bytes to write (or read)
 */
void ssp_wr(struct ssp_t *sspif, uint8 *tx, uint8 *rx, uint32 len)
{
    /* sanity check */
    if (len == 0)
        return;

    while (len--)
    {
        if (tx != NULL)
            sspif->pm->DR = *tx++;
        else
            sspif->pm->DR = 0xff;

        while ((sspif->pm->SR & (SSP_SR_RNE | SSP_SR_BSY)) != SSP_SR_RNE)
            ;

        if (rx != NULL)
            *rx++ = sspif->pm->DR;
        else
            sspif->pm->DR;
    }
}

/*
 *
 */
void ssp_disable(struct ssp_t *sspif)
{
    /* disable SSP */
    sspif->pm->CR1 &= ~SSP_CR1_SSE;

    if (sspif == SSP0)
    {
        LPC_IOCON_FUNC_SET(P1_20, 0);
        LPC_IOCON_FUNC_SET(P1_23, 0);
        LPC_IOCON_FUNC_SET(P1_24, 0);
    } else if (sspif == SSP1) {
        LPC_IOCON_FUNC_SET(P0_7, 0);
        LPC_IOCON_FUNC_SET(P0_8, 0);
        LPC_IOCON_FUNC_SET(P0_9, 0);
    } else {
        /* NOTREACHED */
        debug_emsg("unknown SSP interface");
        return;
    }
}

