#!/bin/sh
#\
exec tclsh "$0" "$@"

#####################################################################
# Dumb converter from BDF to some raw format, not fully compliant to
# "Glyph Bitmap Distribution Format (BDF) Specification" from Adobe.
#
# NOTE Only fixed-width fonts are supported
#
# Some useful font links:
#     http://terminus-font.sourceforge.net
#     http://www.cl.cam.ac.uk/~mgk25/ucs-fonts.html
#     http://unifoundry.com/unifont.html
#
#####################################################################

if {$argc < 2} {
    puts "Usage:"
    puts "    [file tail [info script]] \[INPUT_FILE\] \[OUTPUT_FILE\] <PAGES>"
    puts ""
    puts "    INPUT_FILE     should be in BDF format"
    puts "    OUTPUT_FILE    can be one of \".pbm\" or \".bin\" format"
    puts "    PAGES          UCS-2 \"page\" list or \"all\" for all pages"
    puts ""
    exit 1
}

set INPUT  [lindex $argv 0]
set OUTPUT [lindex $argv 1]
set PAGES  [lrange $argv 2 end]

if {[llength $PAGES] == 0} {
    set PAGES [list 0x00]
}

if {[lindex $PAGES 0] eq "all"} {
    set PAGES [list]
    for {set i 0} {$i < 256} {incr i} {
        lappend PAGES $i
    }
}

if {[file extension $INPUT] ne ".bdf"} {
    puts stderr {Invalid input file extension}
    exit 1
}

switch [file extension $OUTPUT] {
    .pbm {set FORMAT PBM}
    .bin {set FORMAT BIN}
    default {
        puts stderr {Invalid output file extension}
        exit 1
    }
}

#
#
#
proc getchars {file} {
    set chars [list]
    set fd [open $file r]

    set info [dict create]

    while {[gets $fd line] > 0} {

        if {![info exists char]} {
            if {[regexp {^SIZE\s+([\d]+)\s+([\d]+)\s+([\d]+)} $line -> PointSize Xres Yres]} {
                dict set info SIZE [dict create {*}[list         \
                                            PointSize $PointSize \
                                            Xres      $Xres      \
                                            Yres      $Yres      \
                                            ]]
            }
            if {[regexp {^PIXEL_SIZE\s+([\d]+)} $line -> size]} {
                dict set info PIXEL_SIZE $size
            }
            if {[regexp {^STARTCHAR\s+([\d\w]+)} $line -> name]} {
                set char [dict create]
                dict set char CHARNAME $name
                set catchbitmap 0
                set bitmap [list]
            }
        } else {
            if {!$catchbitmap} {
                if {[regexp {^ENCODING\s+([-\d]+)} $line -> encoding]} {
                    if {$encoding < 0} {
                        unset char
                        continue
                    }
                    dict set char ENCODING $encoding
                }
                if {[regexp {^SWITDH\s+([\d]+)\s+([\d]+)} $line -> swx0 swy0]} {
                    dict set char SWIDTH [dict create swx0 $swx0 swy0 $swy0]
                }
                if {[regexp {^DWIDTH\s+([\d]+)\s+([\d]+)} $line -> dwx0 dwy0]} {
                    dict set char DWIDTH [dict create dwx0 $dwx0 dwy0 $dwy0]
                }
                if {[regexp {^BBX\s+([\d]+)\s+([\d]+)\s+([-\d]+)\s+([-\d]+)} $line -> BBw BBh BBxoff0x BByoff0y]} {
                    dict set char BBX [dict create {*}[list        \
                                                BBw      $BBw      \
                                                BBh      $BBh      \
                                                BBxoff0x $BBxoff0x \
                                                BByoff0y $BByoff0y \
                                                ]]
                }
                if {[regexp {^BITMAP} $line]} {
                    set catchbitmap 1
                }
            } else {
                if {[string match {ENDCHAR} $line]} {
                    dict set char BITMAP $bitmap
                    lappend chars $char
                    unset -nocomplain char
                } else {
                    lappend bitmap 0x[string trim $line]
                }
            }
        }
    }

    close $fd
    return [list $info $chars]
}

#
# convert number to binary data
#
proc num2bin {num len} {
    set result {}

    while {$len} {
        append result [binary format c [expr {$num & 0xff}]]
        set num [expr {$num >> 8}]
        incr len -1
    }

    return $result
}

#
#
#
proc selectpage {page chars} {
    set res [list]

    foreach ch $chars {
        if {[dict exists $ch ENCODING]} {
            set encoding [dict get $ch ENCODING]
            if {$encoding >= 0 && ($encoding & 0xff00) == $page} {
                lappend res $ch
            }
        } else {
            puts stderr "Character has not encoding field"
            puts stderr $ch
            exit 1
        }
    }
    return $res
}

lassign [getchars $INPUT] info chars

foreach num $PAGES {
    lappend pages [list $num [selectpage [expr {$num << 8}] $chars]]
}

# XXX get character dimension from first char
if {1} {
    set ch [lindex $chars 0]
    
    set chheight [dict get $info PIXEL_SIZE]
    set chwidth  [dict get [dict get $ch DWIDTH] dwx0]

    set xofs [dict get [dict get $ch BBX] BBxoff0x]
    if {$xofs < 0} {
        set chwidth [expr {$chwidth - $xofs}]
    }

    set bwidth [expr {$chwidth / 8}]
    if {$chwidth % 8} {
        incr bwidth
    }

    set chsize [expr {$bwidth * $chheight}]
}

#puts "dwx0 = $dwx0"
#puts "offs = $BBxoff0x"

#puts "symbols [llength [selectpage 0x0400 $chars]]"

if {$FORMAT eq "BIN"} {
    set fd [open $OUTPUT w]
    fconfigure $fd -translation binary -encoding binary

    if {1} {
        #
        # header format
        #
        #  0 -  1    0xBEEF
        #  2         header version
        #  3 -  4    header length
        #  5         number of pages
        #  6         maximum symbol width in bytes
        #  7         symbol heigth in pixels
        #  8 - 31    reserved
        #  32        page0   index
        #  33        page1   index
        #  ...
        #  xxx       page255 index
        #
        #
        set HEADER_SIZE [expr {32 + 256}]
        append header [num2bin 0xEFBE           2]
        append header [num2bin 0x01             1]
        append header [num2bin $HEADER_SIZE     2]
        append header [num2bin [llength $pages] 1]
        set chw [expr {$chwidth / 8}]
        if {$chwidth % 8} {
            incr chw
        }
        append header [num2bin $chw             1]
        append header [num2bin $chheight        1]
        append header [string repeat \xff [expr {32 - [string length $header]}]]
        for {set n 0; set i 0} {$i < 256} {incr i} {
            set found 0
            foreach page $pages {
                if {$i == [lindex $page 0]} {
                    append header [num2bin $n 1]
                    incr n
                    set found 1
                    break
                }
            }
            if {!$found} {
                append header [num2bin 0 1]
            }
        }
        puts -nonewline $fd $header
    }
    set hoffset [string length $header]

    # Outpu binary format
    # ===================
    #
    # 7x8 character layout:
    #
    #   byte0 --- 6543210
    #   byte1 --- 6543210
    #             6543210
    #             6543210
    #             6543210
    #             6543210
    #             6543210
    #   byte7 --- 6543210
    #
    #     
    # 9x16 character layout:
    #
    #                     --byte0-
    #    ---------------- |      |
    #    |               ||      |
    #  byte1  --- 7654321076543210
    #  byte3  --- 7654321076543210 -- byte2
    #                  ...
    #  byte15 --- 7654321076543210 -- byte14
    # 
    #
    set PAGE_SIZE 256

    if {1} {
        # XXX fill with zero whole file
        seek $fd $hoffset
        puts -nonewline $fd [string repeat \x00 [expr {$PAGE_SIZE * $chsize * [llength $pages]}]]

        set num 0
        foreach p $pages {
            lassign $p -> page
            foreach ch $page {
                set pgpos  [expr {$num * $PAGE_SIZE * ($chsize + 1)}]
                set chpos  [expr {([dict get $ch ENCODING] & 0xff) * ($chsize + 1)}]

                seek $fd [expr {$hoffset + $pgpos + $chpos}]

                set chw  [dict get [dict get $ch DWIDTH] dwx0]
                set xofs [dict get [dict get $ch BBX] BBxoff0x]
                if {$xofs < 0} {
                    set chw [expr {$chw - $xofs}]
                }
                puts -nonewline $fd [num2bin $chw 1]

                foreach word [dict get $ch BITMAP] {
                    set xofs [dict get [dict get $ch BBX] BBxoff0x]
                    if {$xofs > 0} {
                        set word [expr {$word >> abs($xofs)}]
                    } else {
                        set word [expr {$word << abs($xofs)}]
                    }
                    set word [expr {$word >> ($bwidth * 8 - $chwidth)}]
                    puts -nonewline $fd [num2bin $word $bwidth]
                }
            }
            incr num
        }
    }

    close $fd

    exit 0
}


if {$FORMAT eq "PBM"} {
    set fd [open $OUTPUT w]

    puts $fd P1

    set CHAR_PER_COL 16
    set CHAR_PER_ROW 16

    set img_width  [expr {$CHAR_PER_COL * $chwidth}]
    set img_height [expr {$CHAR_PER_ROW * $chheight}]
    puts $fd "[expr {$img_width}] [expr {$img_height * [llength $pages]}]"

    set hoffs [tell $fd]

    # XXX fill with zero character whole file
    set  pos $hoffs
    seek $fd $pos
    set n [expr {$img_height * [llength $pages]}]
    while {$n} {
        # NOTE new line
        puts $fd [string repeat 0 [expr {$img_width}]]
        incr n -1
    }

    set pnum 0
    foreach p $pages {
        lassign $p -> page
        foreach ch $page {
            set idx 0

            foreach word [dict get $ch BITMAP] {
                if {1} {
                    set line {}
                    set n $chwidth
                    set mask 0x01

                    set xofs [dict get [dict get $ch BBX] BBxoff0x]
                    if {$xofs > 0} {
                        set word [expr {$word >> abs($xofs)}]
                    } else {
                        set word [expr {$word << abs($xofs)}]
                    }
                    set word [expr {$word >> ($bwidth * 8 - $chwidth)}]

                    while {$n} {
                        if {$word & $mask} {
                            set line 1[set line]
                        } else {
                            set line 0[set line]
                        }

                        set mask [expr {$mask << 1}]
                        incr n -1
                    }
                }

                set code  [expr {[dict get $ch ENCODING] & 0xff}]
                # NOTE img_width + 1 for new line character
                set  pos $hoffs
                incr pos [expr {$idx * ($img_width + 1)}]
                incr pos [expr {($code % $CHAR_PER_COL) * $chwidth}]
                incr pos [expr {($code / $CHAR_PER_COL) * $chheight * ($img_width + 1)}]
                incr pos [expr {$pnum * ($img_width + 1) * $img_height}]

                seek $fd $pos
                puts -nonewline $fd $line
                incr idx
            }
        }
        incr pnum
    }

    close $fd
    exit 0
}


#
#
# A little guide to UTF-8
# =======================
#
#
# Bits of      First       Last         Bytes in
# code point   code point  code point   sequence  Byte 1      Byte 2      Byte 3      Byte 4      Byte 5      Byte 6
# 7            U+0000      U+007F       1         0xxxxxxx
# 11           U+0080      U+07FF       2         110xxxxx    10xxxxxx
# 16           U+0800      U+FFFF       3         1110xxxx    10xxxxxx    10xxxxxx
# 21           U+10000     U+1FFFFF     4         11110xxx    10xxxxxx    10xxxxxx    10xxxxxx
# 26           U+200000    U+3FFFFFF    5         111110xx    10xxxxxx    10xxxxxx    10xxxxxx    10xxxxxx
# 31           U+4000000   U+7FFFFFFF   6         1111110x    10xxxxxx    10xxxxxx    10xxxxxx    10xxxxxx    10xxxxxx
#
#
# Russian "М"
#
# Unicode code point   0x041C
# UTF8 representation  0x9CD0
#
# 0xD0  11010000
#          -----  MSB
#
# 0x9C  10011100
#         ------  LSB
#
# 0b10.0001.1100
#    4    1    C
#             
#

